@@ -3,7 +3,8 @@
 #ifndef SCM__SCM_H
 #define SCM__SCM_H
 
-/* Copyright (C) 1995,1996,2000,2001, 2002, 2006, 2008, 2009, 2010, 2011 Free Software Foundation, Inc.
+/* Copyright (C) 1995, 1996, 2000, 2001, 2002, 2006, 2008, 2009, 2010,
+ *   2011, 2013, 2014 Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -93,18 +94,17 @@
 #ifdef vms
 # ifndef __GNUC__
 #  include <ssdef.h>
-#   define SCM_SYSCALL(line)                                    \
-  do                                                            \
-    {                                                           \
-      errno = 0;                                                \
-      line;                                                     \
-      if (EVMSERR==errno && (vaxc$errno>>3)==(SS$_CONTROLC>>3)) \
-        {                                                       \
-          SCM_ASYNC_TICK;                                       \
-          continue;                                             \
-        }                                                       \
-    }                                                           \
-  while(0)
+#   define SCM_SYSCALL(line)						\
+  do									\
+    {									\
+      errno = 0;							\
+      line;								\
+      if (EVMSERR == errno && (vaxc$errno>>3)==(SS$_CONTROLC>>3))	\
+        scm_async_tick ();                                              \
+      else								\
+	break;								\
+    }									\
+  while (1)
 # endif /* ndef __GNUC__ */
 #endif /* def vms */
 #endif /* ndef SCM_SYSCALL  */
@@ -112,18 +112,18 @@
 #ifndef SCM_SYSCALL
 # ifdef EINTR
 #  if (EINTR > 0)
-#   define SCM_SYSCALL(line)                    \
-  do                                            \
-    {                                           \
-      errno = 0;                                \
-      line;                                     \
-      if (errno == EINTR)                       \
-        {                                       \
-          SCM_ASYNC_TICK;                       \
-          continue;                             \
-        }                                       \
-    }                                           \
-  while(0)
+#   define SCM_SYSCALL(line)			\
+  do						\
+    {						\
+      errno = 0;				\
+      line;					\
+      if (errno == EINTR)			\
+	{					\
+          scm_async_tick ();                    \
+	  errno = EINTR;			\
+	}					\
+    }						\
+  while (errno == EINTR)
 #  endif /*  (EINTR > 0) */
 # endif /* def EINTR */
 #endif /* ndef SCM_SYSCALL */
@@ -191,6 +191,48 @@
 #define scm_from_off64_t  scm_from_int64
 
 
+
+
+#if defined (vms)
+/* VMS: Implement SCM_I_SETJMP in terms of setjump.  */
+extern int setjump(scm_i_jmp_buf env);
+extern int longjump(scm_i_jmp_buf env, int ret);
+#define SCM_I_SETJMP setjump
+#define SCM_I_LONGJMP longjump
+
+#elif defined (_CRAY1)
+/* Cray: Implement SCM_I_SETJMP in terms of setjump.  */
+extern int setjump(scm_i_jmp_buf env);
+extern int longjump(scm_i_jmp_buf env, int ret);
+#define SCM_I_SETJMP setjump
+#define SCM_I_LONGJMP longjump
+
+#elif defined (__ia64__)
+/* IA64: Implement SCM_I_SETJMP in terms of getcontext. */
+# define SCM_I_SETJMP(JB)			        \
+  ( (JB).fresh = 1,				        \
+    getcontext (&((JB).ctx)),                           \
+    ((JB).fresh ? ((JB).fresh = 0, 0) : 1) )
+# define SCM_I_LONGJMP(JB,VAL) scm_ia64_longjmp (&(JB), VAL)
+void scm_ia64_longjmp (scm_i_jmp_buf *, int);
+
+#else
+/* All other systems just use setjmp and longjmp.  */
+
+#define SCM_I_SETJMP setjmp
+#define SCM_I_LONGJMP longjmp
+#endif
+
+
+
+#if (defined __GNUC__)
+# define SCM_NOINLINE __attribute__ ((__noinline__))
+#else
+# define SCM_NOINLINE /* noinline */
+#endif
+
+
+
 /* The endianness marker in objcode.  */
 #ifdef WORDS_BIGENDIAN
 # define SCM_OBJCODE_ENDIANNESS "BE"
@@ -205,8 +247,8 @@
 #define SCM_OBJCODE_WORD_SIZE  SCM_CPP_STRINGIFY (SIZEOF_VOID_P)
 
 /* Major and minor versions must be single characters. */
-#define SCM_OBJCODE_MAJOR_VERSION 2
-#define SCM_OBJCODE_MINOR_VERSION 0
+#define SCM_OBJCODE_MAJOR_VERSION 3
+#define SCM_OBJCODE_MINOR_VERSION 9
 #define SCM_OBJCODE_MAJOR_VERSION_STRING        \
   SCM_CPP_STRINGIFY(SCM_OBJCODE_MAJOR_VERSION)
 #define SCM_OBJCODE_MINOR_VERSION_STRING        \
@@ -216,13 +258,6 @@
 #define SCM_OBJCODE_MACHINE_VERSION_STRING                              \
   SCM_OBJCODE_ENDIANNESS "-" SCM_OBJCODE_WORD_SIZE "-" SCM_OBJCODE_VERSION_STRING
 
-/* The objcode magic header.  */
-#define SCM_OBJCODE_COOKIE                              \
-  "GOOF----" SCM_OBJCODE_MACHINE_VERSION_STRING
-#define SCM_OBJCODE_ENDIANNESS_OFFSET 8
-#define SCM_OBJCODE_WORD_SIZE_OFFSET 11
-
-
 #endif  /* SCM__SCM_H */
 
 /*
