@@ -1,5 +1,5 @@
-/* Copyright (C) 1995, 1996, 1997, 1998, 2000, 2001, 2004, 2006, 2010,
- *   2012, 2013, 2014 Free Software Foundation, Inc.
+/* Copyright (C) 1995-1998, 2000, 2001, 2004, 2006, 2010, 2012-2014
+ *   Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -121,10 +121,13 @@ SCM_DEFINE (scm_strerror, "strerror", 1, 0, 0,
 #define FUNC_NAME s_scm_strerror
 {
   SCM ret;
+  int errnum = scm_to_int (err);  /* Must be done outside of the
+				     critical section below, to avoid a
+				     deadlock on errors.  */
   scm_dynwind_begin (0);
   scm_i_dynwind_pthread_mutex_lock (&scm_i_misc_mutex);
 
-  ret = scm_from_locale_string (strerror (scm_to_int (err)));
+  ret = scm_from_locale_string (strerror (errnum));
 
   scm_dynwind_end ();
   return ret;
@@ -160,7 +163,7 @@ scm_syserror (const char *subr)
   */
 #ifdef EINTR
   if (scm_to_int (err) == EINTR)
-    SCM_ASYNC_TICK;
+    scm_async_tick ();
 #endif
 
   scm_error (scm_system_error_key,
@@ -176,7 +179,7 @@ scm_syserror_msg (const char *subr, const char *message, SCM args, int eno)
   /* See above note about the EINTR signal handling race. */
 #ifdef EINTR
   if (eno == EINTR)
-    SCM_ASYNC_TICK;
+    scm_async_tick ();
 #endif
   scm_error (scm_system_error_key,
 	     subr,
@@ -289,14 +292,6 @@ scm_wrong_type_arg_msg (const char *subr, int pos, SCM bad_value, const char *sz
 }
 
 
-SCM_GLOBAL_SYMBOL (scm_memory_alloc_key, "memory-allocation-error");
-void
-scm_memory_error (const char *subr)
-{
-  fprintf (stderr, "FATAL: memory error in %s\n", subr);
-  abort ();
-}
-
 SCM_GLOBAL_SYMBOL (scm_misc_error_key, "misc-error");
 void
 scm_misc_error (const char *subr, const char *message, SCM args)
