@@ -307,29 +307,24 @@ block_self (SCM queue, scm_i_pthread_mutex_t *mutex,
   SCM q_handle;
   int err;
 
-  if (scm_i_setup_sleep (t, mutex, -1))
-    {
-      scm_i_reset_sleep (t);
-      err = EINTR;
-    }
-  else
-    {
-      t->block_asyncs++;
-      q_handle = enqueue (queue, t->handle);
-      if (waittime == NULL)
-	err = scm_i_scm_pthread_cond_wait (&t->sleep_cond, mutex);
-      else
-	err = scm_i_scm_pthread_cond_timedwait (&t->sleep_cond, mutex, waittime);
+  if (scm_i_prepare_to_wait_on_cond (t, mutex, &t->sleep_cond))
+    return EINTR;
 
-      /* When we are still on QUEUE, we have been interrupted.  We
-	 report this only when no other error (such as a timeout) has
-	 happened above.
-      */
-      if (remqueue (queue, q_handle) && err == 0)
-	err = EINTR;
-      t->block_asyncs--;
-      scm_i_reset_sleep (t);
-    }
+  t->block_asyncs++;
+  q_handle = enqueue (queue, t->handle);
+  if (waittime == NULL)
+    err = scm_i_scm_pthread_cond_wait (&t->sleep_cond, mutex);
+  else
+    err = scm_i_scm_pthread_cond_timedwait (&t->sleep_cond, mutex, waittime);
+
+  /* When we are still on QUEUE, we have been interrupted.  We
+     report this only when no other error (such as a timeout) has
+     happened above.
+  */
+  if (remqueue (queue, q_handle) && err == 0)
+    err = EINTR;
+  t->block_asyncs--;
+  scm_i_wait_finished (t);
 
   return err;
 }
@@ -377,7 +372,7 @@ static SCM default_dynamic_state;
 /* Perform first stage of thread initialisation, in non-guile mode.
  */
 static void
-guilify_self_1 (struct GC_stack_base *base)
+guilify_self_1 (struct GC_stack_base *base, int needs_unregister)
 {
   scm_i_thread t;
 
@@ -413,9 +408,9 @@ guilify_self_1 (struct GC_stack_base *base)
        currently have type `void'.  */
     abort ();
 
-  scm_i_pthread_mutex_init (&t.admin_mutex, NULL);
   t.exited = 0;
   t.guile_mode = 0;
+  t.needs_unregister = needs_unregister;
 
   /* The switcheroo.  */
   {
@@ -463,6 +458,7 @@ guilify_self_2 (SCM dynamic_state)
   }
 
   t->dynamic_state = scm_gc_typed_calloc (scm_t_dynamic_state);
+  t->dynamic_state->thread_local_values = scm_c_make_hash_table (0);
   scm_set_current_dynamic_state (dynamic_state);
 
   t->dynstack.base = scm_gc_malloc (16 * sizeof (scm_t_bits), "dynstack");
@@ -481,7 +477,9 @@ guilify_self_2 (SCM dynamic_state)
 static void
 on_thread_exit (void *v)
 {
-  /* This handler is executed in non-guile mode.  */
+  /* This handler is executed in non-guile mode.  Note that although
+     libgc isn't guaranteed to see thread-locals, for this thread-local
+     that isn't an issue as we have the all_threads list.  */
   scm_i_thread *t = (scm_i_thread *) v, **tp;
 
   t->exited = 1;
@@ -512,14 +510,28 @@ on_thread_exit (void *v)
 
   scm_i_pthread_mutex_unlock (&thread_admin_mutex);
 
-  if (t->vp)
-    {
-      scm_i_vm_free_stack (t->vp);
-      t->vp = NULL;
-    }
+  /* Although this thread has exited, the thread object might still be
+     alive.  Release unused memory.  */
+  t->freelists = NULL;
+  t->pointerless_freelists = NULL;
+  t->dynamic_state = NULL;
+  t->dynstack.base = NULL;
+  t->dynstack.top = NULL;
+  t->dynstack.limit = NULL;
+  {
+    struct scm_vm *vp = t->vp;
+    t->vp = NULL;
+    if (vp)
+      scm_i_vm_free_stack (vp);
+  }
+
+#ifdef SCM_HAVE_THREAD_STORAGE_CLASS
+  scm_i_current_thread = NULL;
+#endif
 
 #if SCM_USE_PTHREAD_THREADS
-  GC_unregister_my_thread ();
+  if (t->needs_unregister)
+    GC_unregister_my_thread ();
 #endif
 }
 
@@ -581,6 +593,8 @@ scm_i_init_thread_for_guile (struct GC_stack_base *base,
 	}
       else
 	{
+          int needs_unregister = 0;
+
 	  /* Guile is already initialized, but this thread enters it for
 	     the first time.  Only initialize this thread.
 	  */
@@ -588,10 +602,11 @@ scm_i_init_thread_for_guile (struct GC_stack_base *base,
 
           /* Register this thread with libgc.  */
 #if SCM_USE_PTHREAD_THREADS
-          GC_register_my_thread (base);
+          if (GC_register_my_thread (base) == GC_SUCCESS)
+            needs_unregister = 1;
 #endif
 
-	  guilify_self_1 (base);
+	  guilify_self_1 (base, needs_unregister);
 	  guilify_self_2 (dynamic_state);
 	}
       return 1;
@@ -730,14 +745,51 @@ scm_call_with_new_thread (SCM thunk, SCM handler)
   return scm_call_2 (call_with_new_thread, thunk, handler);
 }
 
-typedef struct {
+typedef struct launch_data launch_data;
+
+struct launch_data {
+  launch_data *prev;
+  launch_data *next;
   SCM dynamic_state;
   SCM thunk;
-} launch_data;
+};
+
+/* GC-protect the launch data for new threads.  */
+static launch_data *protected_launch_data;
+static scm_i_pthread_mutex_t protected_launch_data_lock =
+  SCM_I_PTHREAD_MUTEX_INITIALIZER;
+
+static void
+protect_launch_data (launch_data *data)
+{
+  scm_i_pthread_mutex_lock (&protected_launch_data_lock);
+  data->next = protected_launch_data;
+  if (protected_launch_data)
+    protected_launch_data->prev = data;
+  protected_launch_data = data;
+  scm_i_pthread_mutex_unlock (&protected_launch_data_lock);
+}
+
+static void
+unprotect_launch_data (launch_data *data)
+{
+  scm_i_pthread_mutex_lock (&protected_launch_data_lock);
+  if (data->next)
+    data->next->prev = data->prev;
+  if (data->prev)
+    data->prev->next = data->next;
+  else
+    protected_launch_data = data->next;
+  scm_i_pthread_mutex_unlock (&protected_launch_data_lock);
+}
 
 static void *
 really_launch (void *d)
 {
+  scm_i_thread *t = SCM_I_CURRENT_THREAD;
+  unprotect_launch_data (d);
+  /* The thread starts with asyncs blocked.  */
+  t->block_asyncs++;
   SCM_I_CURRENT_THREAD->result = scm_call_0 (((launch_data *)d)->thunk);
   return 0;
 }
@@ -766,6 +818,7 @@ SCM_DEFINE (scm_sys_call_with_new_thread, "%call-with-new-thread", 1, 0, 0,
   data = scm_gc_typed_calloc (launch_data);
   data->dynamic_state = scm_current_dynamic_state ();
   data->thunk = thunk;
+  protect_launch_data (data);
   err = scm_i_pthread_create (&id, NULL, launch_thread, data);
   if (err)
     {
@@ -1293,7 +1346,23 @@ timed_wait (enum scm_mutex_kind kind, struct scm_mutex *m, struct scm_cond *c,
       err = block_self (c->waiting, &m->lock, waittime);
 
       /* We woke up for some reason.  Reacquire the mutex before doing
-         anything else.  */
+         anything else.
+
+         FIXME: We disable interrupts while reacquiring the mutex.  If
+         we allow interrupts here, there's the risk of a nonlocal exit
+         before we reaquire the mutex, which would be visible to user
+         code.
+
+         For example the unwind handler in
+
+           (with-mutex m (wait-condition-variable c m))
+
+         that tries to unlock M could see M in an already-unlocked
+         state, if an interrupt while waiting on C caused the wait to
+         abort and the woke thread lost the race to reacquire M.  That's
+         not great.  Maybe it's necessary but for now we just disable
+         interrupts while reaquiring a mutex after a wait.  */
+      current_thread->block_asyncs++;
       if (kind == SCM_MUTEX_RECURSIVE &&
           scm_is_eq (m->owner, current_thread->handle))
 	{
@@ -1310,16 +1379,8 @@ timed_wait (enum scm_mutex_kind kind, struct scm_mutex *m, struct scm_cond *c,
                 break;
               }
             block_self (m->waiting, &m->lock, waittime);
-            if (scm_is_eq (m->owner, SCM_BOOL_F))
-              {
-                m->owner = current_thread->handle;
-                scm_i_pthread_mutex_unlock (&m->lock);
-                break;
-              }
-            scm_i_pthread_mutex_unlock (&m->lock);
-            scm_async_tick ();
-            scm_i_scm_pthread_mutex_lock (&m->lock);
           }
+      current_thread->block_asyncs--;
 
       /* Now that we have the mutex again, handle the return value.  */
       if (err == 0)
@@ -1479,43 +1540,45 @@ scm_std_select (int nfds,
       readfds = &my_readfds;
     }
 
-  while (scm_i_setup_sleep (t, NULL, t->sleep_pipe[1]))
+  if (scm_i_prepare_to_wait_on_fd (t, t->sleep_pipe[1]))
     {
-      scm_i_reset_sleep (t);
-      SCM_TICK;
+      eno = EINTR;
+      res = -1;
     }
-
-  wakeup_fd = t->sleep_pipe[0];
-  FD_SET (wakeup_fd, readfds);
-  if (wakeup_fd >= nfds)
-    nfds = wakeup_fd+1;
-
-  args.nfds = nfds;
-  args.read_fds = readfds;
-  args.write_fds = writefds;
-  args.except_fds = exceptfds;
-  args.timeout = timeout;
-
-  /* Explicitly cooperate with the GC.  */
-  scm_without_guile (do_std_select, &args);
-
-  res = args.result;
-  eno = args.errno_value;
-
-  scm_i_reset_sleep (t);
-
-  if (res > 0 && FD_ISSET (wakeup_fd, readfds))
+  else
     {
-      char dummy;
-      full_read (wakeup_fd, &dummy, 1);
+      wakeup_fd = t->sleep_pipe[0];
+      FD_SET (wakeup_fd, readfds);
+      if (wakeup_fd >= nfds)
+        nfds = wakeup_fd+1;
 
-      FD_CLR (wakeup_fd, readfds);
-      res -= 1;
-      if (res == 0)
-	{
-	  eno = EINTR;
-	  res = -1;
-	}
+      args.nfds = nfds;
+      args.read_fds = readfds;
+      args.write_fds = writefds;
+      args.except_fds = exceptfds;
+      args.timeout = timeout;
+
+      /* Explicitly cooperate with the GC.  */
+      scm_without_guile (do_std_select, &args);
+
+      res = args.result;
+      eno = args.errno_value;
+
+      scm_i_wait_finished (t);
+
+      if (res > 0 && FD_ISSET (wakeup_fd, readfds))
+        {
+          char dummy;
+          full_read (wakeup_fd, &dummy, 1);
+
+          FD_CLR (wakeup_fd, readfds);
+          res -= 1;
+          if (res == 0)
+            {
+              eno = EINTR;
+              res = -1;
+            }
+        }
     }
   errno = eno;
   return res;
@@ -1734,7 +1797,7 @@ scm_threads_prehistory (void *base)
 		 GC_MAKE_PROC (GC_new_proc (thread_mark), 0),
 		 0, 1);
 
-  guilify_self_1 ((struct GC_stack_base *) base);
+  guilify_self_1 ((struct GC_stack_base *) base, 0);
 }
 
 scm_t_bits scm_tc16_thread;
